from dataclasses import field
from pydantic.dataclasses import dataclass
from typing import List, Optional

import datetime as dt

DT_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


@dataclass
class Access:
    class Meta:
        name = "ACCESS"

    vector: Optional[str] = field(
        default=None,
        metadata={
            "name": "VECTOR",
            "type": "Element",
        },
    )
    complexity: Optional[str] = field(
        default=None,
        metadata={
            "name": "COMPLEXITY",
            "type": "Element",
        },
    )


@dataclass
class Attack:
    class Meta:
        name = "ATTACK"

    vector: Optional[str] = field(
        default=None,
        metadata={
            "name": "VECTOR",
            "type": "Element",
        },
    )
    complexity: Optional[str] = field(
        default=None,
        metadata={
            "name": "COMPLEXITY",
            "type": "Element",
        },
    )


@dataclass
class AuthTypeList:
    class Meta:
        name = "AUTH_TYPE_LIST"

    auth_type: List[str] = field(
        default_factory=list,
        metadata={
            "name": "AUTH_TYPE",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Base:
    class Meta:
        name = "BASE"

    source: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        },
    )
    value: str = field(
        default="",
        metadata={
            "required": True,
        },
    )


@dataclass
class Bugtraq:
    class Meta:
        name = "BUGTRAQ"

    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "ID",
            "type": "Element",
            "required": True,
        },
    )
    url: Optional[str] = field(
        default=None,
        metadata={
            "name": "URL",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class ChangeLogInfo:
    class Meta:
        name = "CHANGE_LOG_INFO"

    change_date: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "CHANGE_DATE",
            "type": "Element",
            "required": True,
            "format": DT_FORMAT,
        },
    )
    comments: Optional[str] = field(
        default=None,
        metadata={
            "name": "COMMENTS",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class Compliance:
    class Meta:
        name = "COMPLIANCE"

    type_value: Optional[str] = field(
        default=None,
        metadata={
            "name": "TYPE",
            "type": "Element",
            "required": True,
        },
    )
    section: Optional[str] = field(
        default=None,
        metadata={
            "name": "SECTION",
            "type": "Element",
            "required": True,
        },
    )
    description: Optional[str] = field(
        default=None,
        metadata={
            "name": "DESCRIPTION",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class Cve:
    class Meta:
        name = "CVE"

    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "ID",
            "type": "Element",
            "required": True,
        },
    )
    url: Optional[str] = field(
        default=None,
        metadata={
            "name": "URL",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class Explt:
    class Meta:
        name = "EXPLT"

    ref: Optional[str] = field(
        default=None,
        metadata={
            "name": "REF",
            "type": "Element",
            "required": True,
        },
    )
    desc: Optional[str] = field(
        default=None,
        metadata={
            "name": "DESC",
            "type": "Element",
            "required": True,
        },
    )
    link: Optional[str] = field(
        default=None,
        metadata={
            "name": "LINK",
            "type": "Element",
        },
    )


@dataclass
class IdSet:
    class Meta:
        name = "ID_SET"

    id: List[str] = field(
        default_factory=list,
        metadata={
            "name": "ID",
            "type": "Element",
        },
    )
    id_range: List[str] = field(
        default_factory=list,
        metadata={
            "name": "ID_RANGE",
            "type": "Element",
        },
    )


@dataclass
class Impact:
    class Meta:
        name = "IMPACT"

    confidentiality: Optional[str] = field(
        default=None,
        metadata={
            "name": "CONFIDENTIALITY",
            "type": "Element",
        },
    )
    integrity: Optional[str] = field(
        default=None,
        metadata={
            "name": "INTEGRITY",
            "type": "Element",
        },
    )
    availability: Optional[str] = field(
        default=None,
        metadata={
            "name": "AVAILABILITY",
            "type": "Element",
        },
    )


@dataclass
class LastCustomization:
    class Meta:
        name = "LAST_CUSTOMIZATION"

    datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "DATETIME",
            "type": "Element",
            "required": True,
            "format": DT_FORMAT,
        },
    )
    user_login: Optional[str] = field(
        default=None,
        metadata={
            "name": "USER_LOGIN",
            "type": "Element",
        },
    )


@dataclass
class MwInfo:
    class Meta:
        name = "MW_INFO"

    mw_id: Optional[str] = field(
        default=None,
        metadata={
            "name": "MW_ID",
            "type": "Element",
            "required": True,
        },
    )
    mw_type: Optional[str] = field(
        default=None,
        metadata={
            "name": "MW_TYPE",
            "type": "Element",
        },
    )
    mw_platform: Optional[str] = field(
        default=None,
        metadata={
            "name": "MW_PLATFORM",
            "type": "Element",
        },
    )
    mw_alias: Optional[str] = field(
        default=None,
        metadata={
            "name": "MW_ALIAS",
            "type": "Element",
        },
    )
    mw_rating: Optional[str] = field(
        default=None,
        metadata={
            "name": "MW_RATING",
            "type": "Element",
        },
    )
    mw_link: Optional[str] = field(
        default=None,
        metadata={
            "name": "MW_LINK",
            "type": "Element",
        },
    )


@dataclass
class Param:
    class Meta:
        name = "PARAM"

    key: Optional[str] = field(
        default=None,
        metadata={
            "name": "KEY",
            "type": "Element",
            "required": True,
        },
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "name": "VALUE",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class PciReasons:
    class Meta:
        name = "PCI_REASONS"

    pci_reason: List[str] = field(
        default_factory=list,
        metadata={
            "name": "PCI_REASON",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Software:
    class Meta:
        name = "SOFTWARE"

    product: Optional[str] = field(
        default=None,
        metadata={
            "name": "PRODUCT",
            "type": "Element",
            "required": True,
        },
    )
    vendor: Optional[str] = field(
        default=None,
        metadata={
            "name": "VENDOR",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class ThreatIntel:
    class Meta:
        name = "THREAT_INTEL"

    id: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    value: str = field(
        default="",
        metadata={
            "required": True,
        },
    )


@dataclass
class VendorReference:
    class Meta:
        name = "VENDOR_REFERENCE"

    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "ID",
            "type": "Element",
            "required": True,
        },
    )
    url: Optional[str] = field(
        default=None,
        metadata={
            "name": "URL",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class Warning:
    class Meta:
        name = "WARNING"

    code: Optional[str] = field(
        default=None,
        metadata={
            "name": "CODE",
            "type": "Element",
        },
    )
    text: Optional[str] = field(
        default=None,
        metadata={
            "name": "TEXT",
            "type": "Element",
            "required": True,
        },
    )
    url: Optional[str] = field(
        default=None,
        metadata={
            "name": "URL",
            "type": "Element",
        },
    )


@dataclass
class BugtraqList:
    class Meta:
        name = "BUGTRAQ_LIST"

    bugtraq: List[Bugtraq] = field(
        default_factory=list,
        metadata={
            "name": "BUGTRAQ",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class ChangeLogList:
    class Meta:
        name = "CHANGE_LOG_LIST"

    change_log_info: List[ChangeLogInfo] = field(
        default_factory=list,
        metadata={
            "name": "CHANGE_LOG_INFO",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class ComplianceList:
    class Meta:
        name = "COMPLIANCE_LIST"

    compliance: List[Compliance] = field(
        default_factory=list,
        metadata={
            "name": "COMPLIANCE",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class CveList:
    class Meta:
        name = "CVE_LIST"

    cve: List[Cve] = field(
        default_factory=list,
        metadata={
            "name": "CVE",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Cvss:
    class Meta:
        name = "CVSS"

    base: Optional[Base] = field(
        default=None,
        metadata={
            "name": "BASE",
            "type": "Element",
        },
    )
    temporal: Optional[str] = field(
        default=None,
        metadata={
            "name": "TEMPORAL",
            "type": "Element",
        },
    )
    vector_string: Optional[str] = field(
        default=None,
        metadata={
            "name": "VECTOR_STRING",
            "type": "Element",
        },
    )
    access: Optional[Access] = field(
        default=None,
        metadata={
            "name": "ACCESS",
            "type": "Element",
        },
    )
    impact: Optional[Impact] = field(
        default=None,
        metadata={
            "name": "IMPACT",
            "type": "Element",
        },
    )
    authentication: Optional[str] = field(
        default=None,
        metadata={
            "name": "AUTHENTICATION",
            "type": "Element",
        },
    )
    exploitability: Optional[str] = field(
        default=None,
        metadata={
            "name": "EXPLOITABILITY",
            "type": "Element",
        },
    )
    remediation_level: Optional[str] = field(
        default=None,
        metadata={
            "name": "REMEDIATION_LEVEL",
            "type": "Element",
        },
    )
    report_confidence: Optional[str] = field(
        default=None,
        metadata={
            "name": "REPORT_CONFIDENCE",
            "type": "Element",
        },
    )


@dataclass
class CvssV3:
    class Meta:
        name = "CVSS_V3"

    base: Optional[Base] = field(
        default=None,
        metadata={
            "name": "BASE",
            "type": "Element",
        },
    )
    temporal: Optional[str] = field(
        default=None,
        metadata={
            "name": "TEMPORAL",
            "type": "Element",
        },
    )
    vector_string: Optional[str] = field(
        default=None,
        metadata={
            "name": "VECTOR_STRING",
            "type": "Element",
        },
    )
    cvss3_version: Optional[str] = field(
        default=None,
        metadata={
            "name": "CVSS3_VERSION",
            "type": "Element",
        },
    )
    attack: Optional[Attack] = field(
        default=None,
        metadata={
            "name": "ATTACK",
            "type": "Element",
        },
    )
    impact: Optional[Impact] = field(
        default=None,
        metadata={
            "name": "IMPACT",
            "type": "Element",
        },
    )
    privileges_required: Optional[str] = field(
        default=None,
        metadata={
            "name": "PRIVILEGES_REQUIRED",
            "type": "Element",
        },
    )
    user_interaction: Optional[str] = field(
        default=None,
        metadata={
            "name": "USER_INTERACTION",
            "type": "Element",
        },
    )
    scope: Optional[str] = field(
        default=None,
        metadata={
            "name": "SCOPE",
            "type": "Element",
        },
    )
    exploit_code_maturity: Optional[str] = field(
        default=None,
        metadata={
            "name": "EXPLOIT_CODE_MATURITY",
            "type": "Element",
        },
    )
    remediation_level: Optional[str] = field(
        default=None,
        metadata={
            "name": "REMEDIATION_LEVEL",
            "type": "Element",
        },
    )
    report_confidence: Optional[str] = field(
        default=None,
        metadata={
            "name": "REPORT_CONFIDENCE",
            "type": "Element",
        },
    )


@dataclass
class Discovery:
    class Meta:
        name = "DISCOVERY"

    remote: Optional[str] = field(
        default=None,
        metadata={
            "name": "REMOTE",
            "type": "Element",
            "required": True,
        },
    )
    auth_type_list: Optional[AuthTypeList] = field(
        default=None,
        metadata={
            "name": "AUTH_TYPE_LIST",
            "type": "Element",
        },
    )
    additional_info: Optional[str] = field(
        default=None,
        metadata={
            "name": "ADDITIONAL_INFO",
            "type": "Element",
        },
    )


@dataclass
class ExpltList:
    class Meta:
        name = "EXPLT_LIST"

    explt: List[Explt] = field(
        default_factory=list,
        metadata={
            "name": "EXPLT",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class MwList:
    class Meta:
        name = "MW_LIST"

    mw_info: List[MwInfo] = field(
        default_factory=list,
        metadata={
            "name": "MW_INFO",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class ParamList:
    class Meta:
        name = "PARAM_LIST"

    param: List[Param] = field(
        default_factory=list,
        metadata={
            "name": "PARAM",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class SoftwareList:
    class Meta:
        name = "SOFTWARE_LIST"

    software: List[Software] = field(
        default_factory=list,
        metadata={
            "name": "SOFTWARE",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class ThreatIntelligence:
    class Meta:
        name = "THREAT_INTELLIGENCE"

    threat_intel: List[ThreatIntel] = field(
        default_factory=list,
        metadata={
            "name": "THREAT_INTEL",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class VendorReferenceList:
    class Meta:
        name = "VENDOR_REFERENCE_LIST"

    vendor_reference: List[VendorReference] = field(
        default_factory=list,
        metadata={
            "name": "VENDOR_REFERENCE",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class ExpltSrc:
    class Meta:
        name = "EXPLT_SRC"

    src_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "SRC_NAME",
            "type": "Element",
            "required": True,
        },
    )
    explt_list: Optional[ExpltList] = field(
        default=None,
        metadata={
            "name": "EXPLT_LIST",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class MwSrc:
    class Meta:
        name = "MW_SRC"

    src_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "SRC_NAME",
            "type": "Element",
            "required": True,
        },
    )
    mw_list: Optional[MwList] = field(
        default=None,
        metadata={
            "name": "MW_LIST",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class Request:
    class Meta:
        name = "REQUEST"

    datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "DATETIME",
            "type": "Element",
            "required": True,
            "format": DT_FORMAT,
        },
    )
    user_login: Optional[str] = field(
        default=None,
        metadata={
            "name": "USER_LOGIN",
            "type": "Element",
            "required": True,
        },
    )
    resource: Optional[str] = field(
        default=None,
        metadata={
            "name": "RESOURCE",
            "type": "Element",
            "required": True,
        },
    )
    param_list: Optional[ParamList] = field(
        default=None,
        metadata={
            "name": "PARAM_LIST",
            "type": "Element",
        },
    )
    post_data: Optional[str] = field(
        default=None,
        metadata={
            "name": "POST_DATA",
            "type": "Element",
        },
    )


@dataclass
class Exploits:
    class Meta:
        name = "EXPLOITS"

    explt_src: List[ExpltSrc] = field(
        default_factory=list,
        metadata={
            "name": "EXPLT_SRC",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Malware:
    class Meta:
        name = "MALWARE"

    mw_src: List[MwSrc] = field(
        default_factory=list,
        metadata={
            "name": "MW_SRC",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Correlation:
    class Meta:
        name = "CORRELATION"

    exploits: Optional[Exploits] = field(
        default=None,
        metadata={
            "name": "EXPLOITS",
            "type": "Element",
        },
    )
    malware: Optional[Malware] = field(
        default=None,
        metadata={
            "name": "MALWARE",
            "type": "Element",
        },
    )


@dataclass
class Vuln:
    class Meta:
        name = "VULN"

    qid: Optional[int] = field(
        default=None,
        metadata={
            "name": "QID",
            "type": "Element",
            "required": True,
        },
    )
    vuln_type: Optional[str] = field(
        default=None,
        metadata={
            "name": "VULN_TYPE",
            "type": "Element",
            "required": True,
        },
    )
    severity_level: Optional[int] = field(
        default=None,
        metadata={
            "name": "SEVERITY_LEVEL",
            "type": "Element",
            "required": True,
        },
    )
    title: Optional[str] = field(
        default=None,
        metadata={
            "name": "TITLE",
            "type": "Element",
            "required": True,
        },
    )
    category: Optional[str] = field(
        default=None,
        metadata={
            "name": "CATEGORY",
            "type": "Element",
        },
    )
    detection_info: Optional[str] = field(
        default=None,
        metadata={
            "name": "DETECTION_INFO",
            "type": "Element",
        },
    )
    last_customization: Optional[LastCustomization] = field(
        default=None,
        metadata={
            "name": "LAST_CUSTOMIZATION",
            "type": "Element",
        },
    )
    last_service_modification_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_SERVICE_MODIFICATION_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    published_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "PUBLISHED_DATETIME",
            "type": "Element",
            "required": True,
            "format": DT_FORMAT,
        },
    )
    bugtraq_list: Optional[BugtraqList] = field(
        default=None,
        metadata={
            "name": "BUGTRAQ_LIST",
            "type": "Element",
        },
    )
    patchable: Optional[str] = field(
        default=None,
        metadata={
            "name": "PATCHABLE",
            "type": "Element",
            "required": True,
        },
    )
    software_list: Optional[SoftwareList] = field(
        default=None,
        metadata={
            "name": "SOFTWARE_LIST",
            "type": "Element",
        },
    )
    vendor_reference_list: Optional[VendorReferenceList] = field(
        default=None,
        metadata={
            "name": "VENDOR_REFERENCE_LIST",
            "type": "Element",
        },
    )
    cve_list: Optional[CveList] = field(
        default=None,
        metadata={
            "name": "CVE_LIST",
            "type": "Element",
        },
    )
    diagnosis: Optional[str] = field(
        default=None,
        metadata={
            "name": "DIAGNOSIS",
            "type": "Element",
        },
    )
    diagnosis_comment: Optional[str] = field(
        default=None,
        metadata={
            "name": "DIAGNOSIS_COMMENT",
            "type": "Element",
        },
    )
    consequence: Optional[str] = field(
        default=None,
        metadata={
            "name": "CONSEQUENCE",
            "type": "Element",
        },
    )
    consequence_comment: Optional[str] = field(
        default=None,
        metadata={
            "name": "CONSEQUENCE_COMMENT",
            "type": "Element",
        },
    )
    solution: Optional[str] = field(
        default=None,
        metadata={
            "name": "SOLUTION",
            "type": "Element",
        },
    )
    solution_comment: Optional[str] = field(
        default=None,
        metadata={
            "name": "SOLUTION_COMMENT",
            "type": "Element",
        },
    )
    compliance_list: Optional[ComplianceList] = field(
        default=None,
        metadata={
            "name": "COMPLIANCE_LIST",
            "type": "Element",
        },
    )
    correlation: Optional[Correlation] = field(
        default=None,
        metadata={
            "name": "CORRELATION",
            "type": "Element",
        },
    )
    cvss: Optional[Cvss] = field(
        default=None,
        metadata={
            "name": "CVSS",
            "type": "Element",
        },
    )
    cvss_v3: Optional[CvssV3] = field(
        default=None,
        metadata={
            "name": "CVSS_V3",
            "type": "Element",
        },
    )
    pci_flag: Optional[str] = field(
        default=None,
        metadata={
            "name": "PCI_FLAG",
            "type": "Element",
        },
    )
    automatic_pci_fail: Optional[str] = field(
        default=None,
        metadata={
            "name": "AUTOMATIC_PCI_FAIL",
            "type": "Element",
        },
    )
    pci_reasons: Optional[PciReasons] = field(
        default=None,
        metadata={
            "name": "PCI_REASONS",
            "type": "Element",
        },
    )
    threat_intelligence: Optional[ThreatIntelligence] = field(
        default=None,
        metadata={
            "name": "THREAT_INTELLIGENCE",
            "type": "Element",
        },
    )
    supported_modules: Optional[str] = field(
        default=None,
        metadata={
            "name": "SUPPORTED_MODULES",
            "type": "Element",
        },
    )
    discovery: Optional[Discovery] = field(
        default=None,
        metadata={
            "name": "DISCOVERY",
            "type": "Element",
            "required": True,
        },
    )
    is_disabled: Optional[str] = field(
        default=None,
        metadata={
            "name": "IS_DISABLED",
            "type": "Element",
        },
    )
    change_log_list: Optional[ChangeLogList] = field(
        default=None,
        metadata={
            "name": "CHANGE_LOG_LIST",
            "type": "Element",
        },
    )


@dataclass
class VulnList:
    class Meta:
        name = "VULN_LIST"

    vuln: List[Vuln] = field(
        default_factory=list,
        metadata={
            "name": "VULN",
            "type": "Element",
        },
    )


@dataclass
class Response:
    class Meta:
        name = "RESPONSE"

    datetime: Optional[str] = field(
        default=None,
        metadata={
            "name": "DATETIME",
            "type": "Element",
            "required": True,
        },
    )
    vuln_list: Optional[VulnList] = field(
        default=None,
        metadata={
            "name": "VULN_LIST",
            "type": "Element",
        },
    )
    id_set: Optional[IdSet] = field(
        default=None,
        metadata={
            "name": "ID_SET",
            "type": "Element",
        },
    )
    warning: Optional[Warning] = field(
        default=None,
        metadata={
            "name": "WARNING",
            "type": "Element",
        },
    )


@dataclass
class KnowledgeBaseVulnListOutput:
    class Meta:
        name = "KNOWLEDGE_BASE_VULN_LIST_OUTPUT"

    request: Optional[Request] = field(
        default=None,
        metadata={
            "name": "REQUEST",
            "type": "Element",
        },
    )
    response: Optional[Response] = field(
        default=None,
        metadata={
            "name": "RESPONSE",
            "type": "Element",
            "required": True,
        },
    )



"""This file contains all the Qualys API URLs used by qualyspy."""

about = "/msp/about.php"
host_list_detection = "/api/2.0/fo/asset/host/vm/detection/"
host_list = "/api/2.0/fo/asset/host/"
knowledgebase = "/api/2.0/fo/knowledge_base/vuln/"

gateway_auth = "/auth/"

asset_details = "/rest/2.0/get/am/asset"
all_asset_details = "/rest/2.0/search/am/asset"



"""Custom xsdata converters to be shared by different models"""

import datetime as dt
from xsdata.formats.converter import Converter
from typing import Any
import ipaddress


class IPv4AddressConverter(Converter):
    """Converter for IPv4Address."""

    def deserialize(
        self, value: str, **kwargs: dict[Any, Any]
    ) -> ipaddress.IPv4Address:
        return ipaddress.IPv4Address(value)

    def serialize(self, value: ipaddress.IPv4Address, **kwargs: dict[Any, Any]) -> str:
        return str(value)


class IPv6AddressConverter(Converter):
    """Converter for IPv6Address."""

    def deserialize(
        self, value: str, **kwargs: dict[Any, Any]
    ) -> ipaddress.IPv6Address:
        return ipaddress.IPv6Address(value)

    def serialize(self, value: ipaddress.IPv6Address, **kwargs: dict[Any, Any]) -> str:
        return str(value)


class TimeDeltaConverter(Converter):
    """Converter for timedelta."""

    def deserialize(self, value: str, **kwargs: dict[Any, Any]) -> dt.timedelta:
        return dt.timedelta(seconds=int(value))

    def serialize(self, value: dt.timedelta, **kwargs: dict[Any, Any]) -> str:
        return str(value.total_seconds())


class StrConverter(Converter):
    """Converter for str, as metadata is a reserved keyword in SQLAlchemy."""

    def deserialize(self, value: str, **kwargs: dict[Any, Any]) -> str:
        if value == "metadata":
            return "metadata_"
        return value

    def serialize(self, value: str, **kwargs: dict[Any, Any]) -> str:
        return str(value)



"""ORM data model for host_list_vm_detection"""

import datetime as dt
import ipaddress
from typing import Optional

import sqlalchemy as sa
import sqlalchemy.orm as orm

from .. import sa_types


class Base(orm.DeclarativeBase):
    pass


Base.metadata.schema = "host_list"


class AssetGroup(Base):
    __tablename__ = "asset_group"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    title: orm.Mapped[str | None]

    asset_group_list_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_group_list.id")
    )
    asset_group_list: orm.Mapped["AssetGroupList"] = orm.relationship(
        "AssetGroupList", back_populates="asset_group"
    )


class Attribute(Base):
    __tablename__ = "attribute"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    name: orm.Mapped[str | None]
    last_status: orm.Mapped[str | None]
    value: orm.Mapped[str | None]
    last_success_date: orm.Mapped[dt.datetime | None]
    last_error_date: orm.Mapped[dt.datetime | None]
    last_error: orm.Mapped[str | None]

    azure_id: orm.Mapped[int | None] = orm.mapped_column(sa.ForeignKey("azure.id"))
    azure: orm.Mapped[Optional["Azure"]] = orm.relationship(back_populates="attribute")
    google_id: orm.Mapped[int | None] = orm.mapped_column(sa.ForeignKey("google.id"))
    google: orm.Mapped[Optional["Google"]] = orm.relationship(
        back_populates="attribute"
    )
    ec2_id: orm.Mapped[int | None] = orm.mapped_column(sa.ForeignKey("ec2.id"))
    ec2: orm.Mapped[Optional["Ec2"]] = orm.relationship(back_populates="attribute")


class CloudTag(Base):
    __tablename__ = "cloud_tag"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    name: orm.Mapped[str | None]
    value: orm.Mapped[str | None]
    last_success_date: orm.Mapped[dt.datetime | None]

    cloud_provider_tags_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("cloud_provider_tags.id")
    )
    cloud_provider_tags: orm.Mapped["CloudProviderTags"] = orm.relationship(
        back_populates="cloud_tag"
    )


class DnsData(Base):
    __tablename__ = "dns_data"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    hostname: orm.Mapped[str | None]
    domain: orm.Mapped[str | None]
    fqdn: orm.Mapped[str | None]

    host_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("host.id"))
    host: orm.Mapped["Host"] = orm.relationship(back_populates="dns_data")


class IdSet(Base):
    __tablename__ = "id_set"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    id_range: orm.Mapped[list[str] | None] = orm.mapped_column(sa.ARRAY(sa.String))


class Param(Base):
    __tablename__ = "param"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    key: orm.Mapped[str | None]
    value: orm.Mapped[str | None]

    param_list_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("param_list.id"))
    param_list: orm.Mapped["ParamList"] = orm.relationship(back_populates="param")


class Tag(Base):
    __tablename__ = "tag"

    tag_id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    name: orm.Mapped[str | None]

    tags_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("tags.id"))
    tags: orm.Mapped["Tags"] = orm.relationship(back_populates="tag")


class User(Base):
    __tablename__ = "user"

    user_login: orm.Mapped[str | None] = orm.mapped_column(primary_key=True)
    first_name: orm.Mapped[str | None]
    last_name: orm.Mapped[str | None]

    user_list_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("user_list.id"))
    user_list: orm.Mapped["UserList"] = orm.relationship(back_populates="user")


class Value1(Base):
    __tablename__ = "value1"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    ud_attr: orm.Mapped[str | None]
    value: orm.Mapped[str | None]

    user_def_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("user_def.id"))
    user_def: orm.Mapped["UserDef"] = orm.relationship(back_populates="value_1")


class Value2(Base):
    __tablename__ = "value2"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    ud_attr: orm.Mapped[str | None]
    value: orm.Mapped[str | None]

    user_def_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("user_def.id"))
    user_def: orm.Mapped["UserDef"] = orm.relationship(back_populates="value_2")


class Value3(Base):
    __tablename__ = "value3"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    ud_attr: orm.Mapped[str | None]
    value: orm.Mapped[str | None]

    user_def_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("user_def.id"))
    user_def: orm.Mapped["UserDef"] = orm.relationship(back_populates="value_3")


class VulnCount(Base):
    __tablename__ = "vuln_count"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    qds_severity: orm.Mapped[int | None]
    value: orm.Mapped[int | None]

    ars_factors_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("ars_factors.id"))
    ars_factors: orm.Mapped["ArsFactors"] = orm.relationship(
        back_populates="vuln_count"
    )


class ArsFactors(Base):
    __tablename__ = "ars_factors"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    ars_formula: orm.Mapped[str | None]
    vuln_count: orm.Mapped[list[VulnCount]] = orm.relationship(
        back_populates="ars_factors"
    )

    host_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("host.id"))
    host: orm.Mapped["Host"] = orm.relationship(back_populates="ars_factors")


class AssetGroupList(Base):
    __tablename__ = "asset_group_list"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    asset_group: orm.Mapped[list[AssetGroup]] = orm.relationship(
        back_populates="asset_group_list"
    )

    glossary_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("glossary.id"))
    glossary: orm.Mapped["Glossary"] = orm.relationship(
        back_populates="asset_group_list"
    )


class Azure(Base):
    __tablename__ = "azure"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    attribute: orm.Mapped[list[Attribute]] = orm.relationship(back_populates="azure")

    metadata_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("metadata.id"))
    metadata_: orm.Mapped["Metadata"] = orm.relationship(back_populates="azure")


class CloudProviderTags(Base):
    __tablename__ = "cloud_provider_tags"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    cloud_tag: orm.Mapped[list[CloudTag]] = orm.relationship(
        back_populates="cloud_provider_tags"
    )

    host_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("host.id"))
    host: orm.Mapped["Host"] = orm.relationship(back_populates="cloud_provider_tags")


class Ec2(Base):
    __tablename__ = "ec2"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    attribute: orm.Mapped[list[Attribute]] = orm.relationship(back_populates="ec2")

    metadata_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("metadata.id"))
    metadata_: orm.Mapped["Metadata"] = orm.relationship(back_populates="ec2")


class Google(Base):
    __tablename__ = "google"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    attribute: orm.Mapped[list[Attribute]] = orm.relationship(back_populates="google")

    metadata_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("metadata.id"))
    metadata_: orm.Mapped["Metadata"] = orm.relationship(back_populates="google")


class ParamList(Base):
    __tablename__ = "param_list"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    param: orm.Mapped[list[Param]] = orm.relationship(back_populates="param_list")


class Tags(Base):
    __tablename__ = "tags"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    tag: orm.Mapped[list[Tag]] = orm.relationship(back_populates="tags")

    host_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("host.id"))
    host: orm.Mapped["Host"] = orm.relationship(back_populates="tags")


class UserDef(Base):
    __tablename__ = "user_def"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    label_1: orm.Mapped[str | None]
    label_2: orm.Mapped[str | None]
    label_3: orm.Mapped[str | None]
    value_1: orm.Mapped[Value1 | None] = orm.relationship(
        back_populates="user_def", uselist=False
    )
    value_2: orm.Mapped[Value2 | None] = orm.relationship(
        back_populates="user_def", uselist=False
    )
    value_3: orm.Mapped[Value3 | None] = orm.relationship(
        back_populates="user_def", uselist=False
    )

    glossary_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("glossary.id"))
    glossary: orm.Mapped["Glossary"] = orm.relationship(back_populates="user_def")

    host_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("host.id"))
    host: orm.Mapped["Host"] = orm.relationship(back_populates="user_def")


class UserList(Base):
    __tablename__ = "user_list"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    user: orm.Mapped[list[User]] = orm.relationship(back_populates="user_list")

    glossary_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("glossary.id"))
    glossary: orm.Mapped["Glossary"] = orm.relationship(back_populates="user_list")


class Glossary(Base):
    __tablename__ = "glossary"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    user_def: orm.Mapped[UserDef] = orm.relationship(
        back_populates="glossary", uselist=False
    )
    user_list: orm.Mapped[UserList] = orm.relationship(
        back_populates="glossary", uselist=False
    )
    asset_group_list: orm.Mapped[AssetGroupList] = orm.relationship(
        back_populates="glossary", uselist=False
    )


class Metadata(Base):
    __tablename__ = "metadata"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    ec2: orm.Mapped[list[Ec2]] = orm.relationship(back_populates="metadata_")
    google: orm.Mapped[list[Google]] = orm.relationship(back_populates="metadata_")
    azure: orm.Mapped[list[Azure]] = orm.relationship(back_populates="metadata_")

    host_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("host.id"))
    host: orm.Mapped["Host"] = orm.relationship(back_populates="metadata_")


class Host(Base):
    __tablename__ = "host"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    asset_id: orm.Mapped[int | None]
    ip: orm.Mapped[ipaddress.IPv4Address | None] = orm.mapped_column(
        "ip", sa_types.IPv4AddressType
    )
    ipv6: orm.Mapped[ipaddress.IPv6Address | None] = orm.mapped_column(
        "ipv6", sa_types.IPv6AddressType
    )
    asset_risk_score: orm.Mapped[int | None]
    asset_criticality_score: orm.Mapped[int | None]
    ars_factors: orm.Mapped[ArsFactors | None] = orm.relationship(
        back_populates="host", uselist=False
    )
    tracking_method: orm.Mapped[str | None]
    network_id: orm.Mapped[int | None]
    dns: orm.Mapped[str | None]
    dns_data: orm.Mapped[DnsData | None] = orm.relationship(
        back_populates="host", uselist=False
    )
    cloud_provider: orm.Mapped[str | None]
    cloud_service: orm.Mapped[str | None]
    cloud_resource_id: orm.Mapped[str | None]
    ec2_instance_id: orm.Mapped[str | None]
    netbios: orm.Mapped[str | None]
    os: orm.Mapped[str | None]
    qg_hostid: orm.Mapped[str | None]
    tags: orm.Mapped[Tags | None] = orm.relationship(
        back_populates="host", uselist=False
    )
    metadata_: orm.Mapped[Metadata | None] = orm.relationship(
        back_populates="host", uselist=False
    )
    cloud_provider_tags: orm.Mapped[CloudProviderTags | None] = orm.relationship(
        back_populates="host", uselist=False
    )
    last_vuln_scan_datetime: orm.Mapped[dt.datetime | None]
    last_vm_scanned_date: orm.Mapped[dt.datetime | None]
    last_vm_scanned_duration: orm.Mapped[dt.timedelta | None]
    last_vm_auth_scanned_date: orm.Mapped[dt.datetime | None]
    last_vm_auth_scanned_duration: orm.Mapped[dt.timedelta | None]
    last_compliance_scan_datetime: orm.Mapped[dt.datetime | None]
    last_scap_scan_datetime: orm.Mapped[dt.datetime | None]
    owner: orm.Mapped[str | None]
    comments: orm.Mapped[str | None]
    user_def: orm.Mapped[UserDef | None] = orm.relationship(back_populates="host", uselist=False)
    asset_group_ids: orm.Mapped[str | None]



# mypy: ignore-errors
# type: ignore
# pylama: ignore=E402

import inspect
import ipaddress
import os
import sys
import unittest

import sqlalchemy.orm as orm

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)


import qualyspy.models.vmdr.host_list_orm as host_list_orm
import qualyspy.models.vmdr.host_list_vm_detection_orm as host_list_vm_detection_orm
from qualyspy import qutils, vmdr
from tests import test_data


class TestQutils(unittest.TestCase):
    def test_to_orm_object(self):
        data = test_data.example_hostlist
        orm_object = qutils.to_orm_object(data.host[0], host_list_vm_detection_orm.Host)
        self.assertEqual(orm_object.ip, data.host[0].ip)


class TestVMDR(unittest.TestCase):
    def test_host_list_detection(self):
        api = vmdr.VmdrAPI()
        hostlist_detection, _, _ = api.host_list_detection(
            ids=test_data.test_host_list_id
        )
        ip = hostlist_detection.host[0].ip
        self.assertEqual(ip, ipaddress.ip_address(test_data.test_host_list_ip_address))

    def test_vmdr_host_list_det_orm_load(self):
        vmdr_orm = vmdr.HostListDetectionORM(echo=True)
        vmdr_orm.load()
        with orm.Session(vmdr_orm.engine) as session:
            host_list = (
                session.query(host_list_vm_detection_orm.Host)
                .where(host_list_vm_detection_orm.Host.id == 11619472)
                .all()
            )
            self.assertEqual(host_list[0].ip, ipaddress.ip_address("172.16.76.84"))

    def test_vmdr_host_list(self):
        api = vmdr.VmdrAPI()
        host_list, _, _ = api.host_list(ids=test_data.test_host_list_id)
        ip = host_list.host[0].ip
        self.assertEqual(ip, ipaddress.ip_address(test_data.test_host_list_ip_address))

    def test_vmdr_host_list_orm_load(self):
        vmdr_orm = vmdr.HostListORM(echo=True)
        vmdr_orm.load()
        with orm.Session(vmdr_orm.engine) as session:
            host_list = (
                session.query(host_list_orm.Host)
                .where(host_list_orm.Host.id == 11619472)
                .all()
            )
            self.assertEqual(host_list[0].ip, ipaddress.ip_address("172.16.76.84"))


class TestKnowledgebase(unittest.TestCase):
    def test_vmdr_knowledgebase(self):
        api = vmdr.VmdrAPI()
        knowledgebase = api.knowledgebase(ids=105943)
        print("test")
        self.assertEqual(knowledgebase.vuln[0].qid, 105943)


if __name__ == "__main__":
    unittest.main()



# generated by datamodel-codegen:
#   filename:  qualys_asset_response.json
#   timestamp: 2023-06-20T16:00:10+00:00

# ([a-z]+)([A-Z])([a-z]+)
# $1_\l$2$3

from __future__ import annotations

import datetime as dt
import ipaddress

from pydantic import BaseModel
from pydantic.utils import to_lower_camel


class Model(BaseModel):
    class Config:
        allow_population_by_field_name = True
        alias_generator = to_lower_camel


class Taxonomy(Model):
    id: int | None
    name: str
    category1: str
    category2: str


class OperatingSystem(Model):
    os_name: str
    full_name: str
    category: str
    category1: str
    category2: str
    product_name: str
    publisher: str
    edition: str | None
    market_version: str | None
    version: str | None
    update: str | None
    architecture: str | None
    lifecycle: str | None
    taxonomy: Taxonomy
    product_url: str | None
    product_family: str | None
    install_date: str | None
    release: str


class Hardware(Model):
    full_name: str | None
    category: str
    category1: str
    category2: str
    manufacturer: str
    product_name: str
    model: str | None
    lifecycle: str | None
    taxonomy: Taxonomy
    product_url: str | None
    product_family: str | None


class UserAccountItem(Model):
    name: str


class UserAccountListData(Model):
    user_account: list[UserAccountItem]


class OpenPortItem(Model):
    port: int
    description: str
    protocol: str
    detected_service: str | None
    first_found: dt.datetime | None
    last_updated: dt.datetime | None


class OpenPortListData(Model):
    open_port: list[OpenPortItem]


class VolumeItem(Model):
    name: str
    free: int
    size: int


class VolumeListData(Model):
    volume: list[VolumeItem]


class NetworkInterfaceItem(Model):
    hostname: str | None
    address_ip_v4: list[ipaddress.IPv4Address] | None
    address_ip_v6: list[ipaddress.IPv6Address] | None
    mac_address: str
    interface_name: str
    dns_address: str | None
    gateway_address: str
    manufacturer: str | None
    mac_vendor_intro_date: dt.datetime | None
    netmask: str | None
    addresses: str | None


class NetworkInterfaceListData(Model):
    network_interface: list[NetworkInterfaceItem]


class SoftwareItem(Model):
    id: int
    full_name: str
    software_type: str
    is_ignored: bool
    ignored_reason: str | None
    category: str | None
    category1: str | None
    category2: str | None
    product_name: str | None
    component: str | None
    publisher: str | None
    edition: str | None
    market_version: str | None
    version: str | None
    update: str | None
    architecture: str | None
    install_date: dt.datetime | None
    install_path: str | None
    last_updated: dt.datetime | None
    last_use_date: dt.datetime | None
    language: str | None
    formerly_known_as: str | None
    is_package: bool | None
    is_package_component: bool | None
    package_name: str | None
    product_url: str | None
    lifecycle: str | None
    support_stage_desc: str | None
    license: str | None
    authorization: str | None


class SoftwareListData(Model):
    software: list[SoftwareItem]


class Activation(Model):
    key: str
    status: str


class Oci_Compute(Model):
    oci_id: str
    tenant_id: str | None
    tenant_name: str | None
    compartment_id: str | None
    compartment_name: str | None
    image: str | None
    shape: str | None
    state: str
    region: str | None
    availability_domain: str | None
    fault_domain: str | None
    creation_date: dt.datetime
    has_agent: bool
    qualys_scanner: bool


class CloudTag(Model):
    key: str
    value: str
    type: str
    name_space: str


class Oci(Model):
    compute: Oci_Compute
    tags: list[CloudTag]


class CloudProvider(Model):
    oci: Oci | None


class Agent(Model):
    version: str
    configuration_profile: str
    activations: list[Activation]
    connected_from: str
    last_activity: dt.datetime
    last_checked_in: dt.datetime
    last_inventory: dt.datetime
    udc_manifest_assigned: bool
    error_status: bool


class Sensor(Model):
    activated_for_modules: list[str]
    pending_activation_for_modules: list[str] | None
    last_v_m_scan: dt.datetime
    last_compliance_scan: dt.datetime
    last_full_scan: dt.datetime
    last_vm_scan_date_scanner: dt.datetime
    last_vm_scan_date_agent: dt.datetime
    last_pc_scan_date_scanner: dt.datetime
    last_pc_scan_date_agent: dt.datetime
    first_easm_scan_date: dt.datetime | None
    last_easm_scan_date: dt.datetime | None


class Container(Model):
    product: str
    version: str | None
    no_of_containers: int
    no_of_images: int
    has_sensor: bool


class Inventory(Model):
    source: str
    created: dt.datetime
    last_updated: dt.datetime


class Activity(Model):
    source: str
    last_scanned_date: dt.datetime


class TagItem(Model):
    tag_id: int
    tag_name: str
    foreground_color: int
    background_color: int
    business_impact: str | None
    criticality_score: int | None


class TagList(Model):
    tag: list[TagItem]


class ServiceItem(Model):
    description: str | None
    name: str
    status: str


class ServiceList(Model):
    service: list[ServiceItem]


class LastLocation(Model):
    city: str | None
    state: str | None
    country: str
    name: str
    continent: str
    postal: str | None


class Criticality(Model):
    score: int
    is_default: bool
    last_updated: dt.datetime | None


class Processor(Model):
    description: str
    speed: int | None
    numCPUs: int | None
    no_of_socket: int | None
    threads_per_core: int | None
    cores_per_socket: int | None
    multithreading_status: str | None


class AssetItem(Model):
    asset_id: int
    assetUUID: str
    host_id: int
    last_modified_date: dt.datetime
    agent_id: str | None
    created_date: dt.datetime
    sensor_last_updated_date: dt.datetime
    asset_type: str | None
    address: ipaddress.IPv4Address | ipaddress.IPv6Address
    dns_name: str | None
    asset_name: str
    netbios_name: str | None
    time_zone: str | None
    bios_description: str | None
    last_boot: dt.datetime | None
    total_memory: int | None
    cpu_count: int | None
    last_logged_on_user: str | None
    domain_role: str | None
    hwUUID: str | None
    bios_serial_number: str | None
    bios_asset_tag: str | None
    is_container_host: bool
    operating_system: OperatingSystem
    hardware: Hardware
    user_account_list_data: UserAccountListData | None
    open_port_list_data: OpenPortListData | None
    volume_list_data: VolumeListData | None
    network_interface_list_data: NetworkInterfaceListData | None
    software_list_data: SoftwareListData | None
    provider: str | None
    cloud_provider: CloudProvider | None
    agent: Agent | None
    sensor: Sensor
    container: Container | None
    inventory: Inventory
    activity: Activity
    tag_list: TagList | None
    service_list: ServiceList | None
    last_location: LastLocation | None
    criticality: Criticality
    business_information: str | None
    assigned_location: str | None
    business_app_list_data: str | None
    risk_score: int | None
    passive_sensor: str | None
    domain: list[str] | None
    subdomain: list[str] | None
    missing_software: str | None
    whois: list[str] | None
    isp: str | None
    asn: str | None
    easm_tags: str | None
    hosting_category1: str | None
    custom_attributes: str | None
    processor: Processor | None


class AssetListData(Model):
    asset: list[AssetItem]


class AssetDetailsOutput(Model):
    response_message: str
    count: int
    response_code: str
    last_seen_asset_id: int | None
    has_more: int
    asset_list_data: AssetListData



"""Exceptions for QualysPy."""


class QualysAPIError(Exception):
    """Indicates an error with the Qualys API."""

    pass


class ConfigError(Exception):
    """Indicates an error with the configuration file."""

    pass



import datetime as dt
import ipaddress
from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass
from xsdata.formats.converter import converter

from . import converters

converter.register_converter(ipaddress.IPv4Address, converters.IPv4AddressConverter())
converter.register_converter(ipaddress.IPv6Address, converters.IPv6AddressConverter())
converter.register_converter(dt.timedelta, converters.TimeDeltaConverter())
converter.register_converter(str, converters.StrConverter())

DT_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


@dataclass
class AssetGroup:
    class Meta:
        name = "ASSET_GROUP"

    id: Optional[int] = field(
        default=None,
        metadata={
            "name": "ID",
            "type": "Element",
            "required": True,
        },
    )
    title: Optional[str] = field(
        default=None,
        metadata={
            "name": "TITLE",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class Attribute:
    class Meta:
        name = "ATTRIBUTE"

    name: Optional[str] = field(
        default=None,
        metadata={
            "name": "NAME",
            "type": "Element",
            "required": True,
        },
    )
    last_status: Optional[str] = field(
        default=None,
        metadata={
            "name": "LAST_STATUS",
            "type": "Element",
            "required": True,
        },
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "name": "VALUE",
            "type": "Element",
            "required": True,
        },
    )
    last_success_date: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_SUCCESS_DATE",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_error_date: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_ERROR_DATE",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_error: Optional[str] = field(
        default=None,
        metadata={
            "name": "LAST_ERROR",
            "type": "Element",
        },
    )


@dataclass
class CloudTag:
    class Meta:
        name = "CLOUD_TAG"

    name: Optional[str] = field(
        default=None,
        metadata={
            "name": "NAME",
            "type": "Element",
            "required": True,
        },
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "name": "VALUE",
            "type": "Element",
            "required": True,
        },
    )
    last_success_date: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_SUCCESS_DATE",
            "type": "Element",
            "required": True,
            "format": DT_FORMAT,
        },
    )


@dataclass
class DnsData:
    class Meta:
        name = "DNS_DATA"

    hostname: Optional[str] = field(
        default=None,
        metadata={
            "name": "HOSTNAME",
            "type": "Element",
        },
    )
    domain: Optional[str] = field(
        default=None,
        metadata={
            "name": "DOMAIN",
            "type": "Element",
        },
    )
    fqdn: Optional[str] = field(
        default=None,
        metadata={
            "name": "FQDN",
            "type": "Element",
        },
    )


@dataclass
class IdSet:
    class Meta:
        name = "ID_SET"

    id: List[int] = field(
        default_factory=list,
        metadata={
            "name": "ID",
            "type": "Element",
        },
    )
    id_range: List[str] = field(
        default_factory=list,
        metadata={
            "name": "ID_RANGE",
            "type": "Element",
        },
    )


@dataclass
class Param:
    class Meta:
        name = "PARAM"

    key: Optional[str] = field(
        default=None,
        metadata={
            "name": "KEY",
            "type": "Element",
            "required": True,
        },
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "name": "VALUE",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class Tag:
    class Meta:
        name = "TAG"

    tag_id: Optional[int] = field(
        default=None,
        metadata={
            "name": "TAG_ID",
            "type": "Element",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "name": "NAME",
            "type": "Element",
        },
    )


@dataclass
class User:
    class Meta:
        name = "USER"

    user_login: Optional[str] = field(
        default=None,
        metadata={
            "name": "USER_LOGIN",
            "type": "Element",
            "required": True,
        },
    )
    first_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "FIRST_NAME",
            "type": "Element",
            "required": True,
        },
    )
    last_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "LAST_NAME",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class Value1:
    class Meta:
        name = "VALUE_1"

    ud_attr: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    value: str = field(
        default="",
        metadata={
            "required": True,
        },
    )


@dataclass
class Value2:
    class Meta:
        name = "VALUE_2"

    ud_attr: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    value: str = field(
        default="",
        metadata={
            "required": True,
        },
    )


@dataclass
class Value3:
    class Meta:
        name = "VALUE_3"

    ud_attr: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    value: str = field(
        default="",
        metadata={
            "required": True,
        },
    )


@dataclass
class VulnCount:
    class Meta:
        name = "VULN_COUNT"

    qds_severity: Optional[int] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    value: int = field(
        default=0,
        metadata={
            "required": True,
        },
    )


@dataclass
class Warning:
    class Meta:
        name = "WARNING"

    code: Optional[int] = field(
        default=None,
        metadata={
            "name": "CODE",
            "type": "Element",
        },
    )
    text: Optional[str] = field(
        default=None,
        metadata={
            "name": "TEXT",
            "type": "Element",
            "required": True,
        },
    )
    url: Optional[str] = field(
        default=None,
        metadata={
            "name": "URL",
            "type": "Element",
        },
    )


@dataclass
class ArsFactors:
    class Meta:
        name = "ARS_FACTORS"

    ars_formula: Optional[str] = field(
        default=None,
        metadata={
            "name": "ARS_FORMULA",
            "type": "Element",
            "required": True,
        },
    )
    vuln_count: List[VulnCount] = field(
        default_factory=list,
        metadata={
            "name": "VULN_COUNT",
            "type": "Element",
        },
    )


@dataclass
class AssetGroupList:
    class Meta:
        name = "ASSET_GROUP_LIST"

    asset_group: List[AssetGroup] = field(
        default_factory=list,
        metadata={
            "name": "ASSET_GROUP",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Azure:
    class Meta:
        name = "AZURE"

    attribute: List[Attribute] = field(
        default_factory=list,
        metadata={
            "name": "ATTRIBUTE",
            "type": "Element",
        },
    )


@dataclass
class CloudProviderTags:
    class Meta:
        name = "CLOUD_PROVIDER_TAGS"

    cloud_tag: List[CloudTag] = field(
        default_factory=list,
        metadata={
            "name": "CLOUD_TAG",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Ec2:
    class Meta:
        name = "EC2"

    attribute: List[Attribute] = field(
        default_factory=list,
        metadata={
            "name": "ATTRIBUTE",
            "type": "Element",
        },
    )


@dataclass
class Google:
    class Meta:
        name = "GOOGLE"

    attribute: List[Attribute] = field(
        default_factory=list,
        metadata={
            "name": "ATTRIBUTE",
            "type": "Element",
        },
    )


@dataclass
class ParamList:
    class Meta:
        name = "PARAM_LIST"

    param: List[Param] = field(
        default_factory=list,
        metadata={
            "name": "PARAM",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Tags:
    class Meta:
        name = "TAGS"

    tag: List[Tag] = field(
        default_factory=list,
        metadata={
            "name": "TAG",
            "type": "Element",
        },
    )


@dataclass
class UserDef:
    class Meta:
        name = "USER_DEF"

    label_1: Optional[str] = field(
        default=None,
        metadata={
            "name": "LABEL_1",
            "type": "Element",
        },
    )
    label_2: Optional[str] = field(
        default=None,
        metadata={
            "name": "LABEL_2",
            "type": "Element",
        },
    )
    label_3: Optional[str] = field(
        default=None,
        metadata={
            "name": "LABEL_3",
            "type": "Element",
        },
    )
    value_1: Optional[Value1] = field(
        default=None,
        metadata={
            "name": "VALUE_1",
            "type": "Element",
        },
    )
    value_2: Optional[Value2] = field(
        default=None,
        metadata={
            "name": "VALUE_2",
            "type": "Element",
        },
    )
    value_3: Optional[Value3] = field(
        default=None,
        metadata={
            "name": "VALUE_3",
            "type": "Element",
        },
    )


@dataclass
class UserList:
    class Meta:
        name = "USER_LIST"

    user: List[User] = field(
        default_factory=list,
        metadata={
            "name": "USER",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Glossary:
    class Meta:
        name = "GLOSSARY"

    user_def: Optional[UserDef] = field(
        default=None,
        metadata={
            "name": "USER_DEF",
            "type": "Element",
        },
    )
    user_list: Optional[UserList] = field(
        default=None,
        metadata={
            "name": "USER_LIST",
            "type": "Element",
        },
    )
    asset_group_list: Optional[AssetGroupList] = field(
        default=None,
        metadata={
            "name": "ASSET_GROUP_LIST",
            "type": "Element",
        },
    )


@dataclass
class Metadata:
    class Meta:
        name = "METADATA"

    ec2: List[Ec2] = field(
        default_factory=list,
        metadata={
            "name": "EC2",
            "type": "Element",
        },
    )
    google: List[Google] = field(
        default_factory=list,
        metadata={
            "name": "GOOGLE",
            "type": "Element",
        },
    )
    azure: List[Azure] = field(
        default_factory=list,
        metadata={
            "name": "AZURE",
            "type": "Element",
        },
    )


@dataclass
class Request:
    class Meta:
        name = "REQUEST"

    datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "DATETIME",
            "type": "Element",
            "required": True,
            "format": DT_FORMAT,
        },
    )
    user_login: Optional[str] = field(
        default=None,
        metadata={
            "name": "USER_LOGIN",
            "type": "Element",
            "required": True,
        },
    )
    resource: Optional[str] = field(
        default=None,
        metadata={
            "name": "RESOURCE",
            "type": "Element",
            "required": True,
        },
    )
    param_list: Optional[ParamList] = field(
        default=None,
        metadata={
            "name": "PARAM_LIST",
            "type": "Element",
        },
    )
    post_data: Optional[str] = field(
        default=None,
        metadata={
            "name": "POST_DATA",
            "type": "Element",
        },
    )


@dataclass
class Host:
    class Meta:
        name = "HOST"

    id: Optional[int] = field(
        default=None,
        metadata={
            "name": "ID",
            "type": "Element",
            "required": True,
        },
    )
    asset_id: Optional[int] = field(
        default=None,
        metadata={
            "name": "ASSET_ID",
            "type": "Element",
        },
    )
    ip: Optional[ipaddress.IPv4Address] = field(
        default=None,
        metadata={
            "name": "IP",
            "type": "Element",
        },
    )
    ipv6: Optional[ipaddress.IPv6Address] = field(
        default=None,
        metadata={
            "name": "IPV6",
            "type": "Element",
        },
    )
    asset_risk_score: Optional[int] = field(
        default=None,
        metadata={
            "name": "ASSET_RISK_SCORE",
            "type": "Element",
        },
    )
    asset_criticality_score: Optional[int] = field(
        default=None,
        metadata={
            "name": "ASSET_CRITICALITY_SCORE",
            "type": "Element",
        },
    )
    ars_factors: Optional[ArsFactors] = field(
        default=None,
        metadata={
            "name": "ARS_FACTORS",
            "type": "Element",
        },
    )
    tracking_method: Optional[str] = field(
        default=None,
        metadata={
            "name": "TRACKING_METHOD",
            "type": "Element",
        },
    )
    network_id: Optional[int] = field(
        default=None,
        metadata={
            "name": "NETWORK_ID",
            "type": "Element",
        },
    )
    dns: Optional[str] = field(
        default=None,
        metadata={
            "name": "DNS",
            "type": "Element",
        },
    )
    dns_data: Optional[DnsData] = field(
        default=None,
        metadata={
            "name": "DNS_DATA",
            "type": "Element",
        },
    )
    cloud_provider: Optional[str] = field(
        default=None,
        metadata={
            "name": "CLOUD_PROVIDER",
            "type": "Element",
        },
    )
    cloud_service: Optional[str] = field(
        default=None,
        metadata={
            "name": "CLOUD_SERVICE",
            "type": "Element",
        },
    )
    cloud_resource_id: Optional[str] = field(
        default=None,
        metadata={
            "name": "CLOUD_RESOURCE_ID",
            "type": "Element",
        },
    )
    ec2_instance_id: Optional[str] = field(
        default=None,
        metadata={
            "name": "EC2_INSTANCE_ID",
            "type": "Element",
        },
    )
    netbios: Optional[str] = field(
        default=None,
        metadata={
            "name": "NETBIOS",
            "type": "Element",
        },
    )
    os: Optional[str] = field(
        default=None,
        metadata={
            "name": "OS",
            "type": "Element",
        },
    )
    qg_hostid: Optional[str] = field(
        default=None,
        metadata={
            "name": "QG_HOSTID",
            "type": "Element",
        },
    )
    tags: Optional[Tags] = field(
        default=None,
        metadata={
            "name": "TAGS",
            "type": "Element",
        },
    )
    metadata: Optional[Metadata] = field(
        default=None,
        metadata={
            "name": "METADATA",
            "type": "Element",
        },
    )
    cloud_provider_tags: Optional[CloudProviderTags] = field(
        default=None,
        metadata={
            "name": "CLOUD_PROVIDER_TAGS",
            "type": "Element",
        },
    )
    last_vuln_scan_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_VULN_SCAN_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_vm_scanned_date: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_VM_SCANNED_DATE",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_vm_scanned_duration: Optional[dt.timedelta] = field(
        default=None,
        metadata={
            "name": "LAST_VM_SCANNED_DURATION",
            "type": "Element",
        },
    )
    last_vm_auth_scanned_date: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_VM_AUTH_SCANNED_DATE",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_vm_auth_scanned_duration: Optional[dt.timedelta] = field(
        default=None,
        metadata={
            "name": "LAST_VM_AUTH_SCANNED_DURATION",
            "type": "Element",
        },
    )
    last_compliance_scan_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_COMPLIANCE_SCAN_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_scap_scan_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_SCAP_SCAN_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    owner: Optional[str] = field(
        default=None,
        metadata={
            "name": "OWNER",
            "type": "Element",
        },
    )
    comments: Optional[str] = field(
        default=None,
        metadata={
            "name": "COMMENTS",
            "type": "Element",
        },
    )
    user_def: Optional[UserDef] = field(
        default=None,
        metadata={
            "name": "USER_DEF",
            "type": "Element",
        },
    )
    asset_group_ids: Optional[str] = field(
        default=None,
        metadata={
            "name": "ASSET_GROUP_IDS",
            "type": "Element",
        },
    )


@dataclass
class HostList:
    class Meta:
        name = "HOST_LIST"

    host: List[Host] = field(
        default_factory=list,
        metadata={
            "name": "HOST",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Response:
    class Meta:
        name = "RESPONSE"

    datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "DATETIME",
            "type": "Element",
            "required": True,
            "format": DT_FORMAT,
        },
    )
    host_list: Optional[HostList] = field(
        default=None,
        metadata={
            "name": "HOST_LIST",
            "type": "Element",
        },
    )
    id_set: Optional[IdSet] = field(
        default=None,
        metadata={
            "name": "ID_SET",
            "type": "Element",
        },
    )
    warning: Optional[Warning] = field(
        default=None,
        metadata={
            "name": "WARNING",
            "type": "Element",
        },
    )
    glossary: Optional[Glossary] = field(
        default=None,
        metadata={
            "name": "GLOSSARY",
            "type": "Element",
        },
    )


@dataclass
class HostListOutput:
    class Meta:
        name = "HOST_LIST_OUTPUT"

    request: Optional[Request] = field(
        default=None,
        metadata={
            "name": "REQUEST",
            "type": "Element",
        },
    )
    response: Optional[Response] = field(
        default=None,
        metadata={
            "name": "RESPONSE",
            "type": "Element",
            "required": True,
        },
    )



# mypy: ignore-errors
# type: ignore
# pylama: ignore=E402

import inspect
import ipaddress
import os
import sys
import unittest

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)

from qualyspy import gav

import sqlalchemy.orm as orm
import qualyspy.models.gav.asset_details_orm as asset_details_orm


class TestGAV(unittest.TestCase):
    def test_gav_asset_details(self):
        api = gav.GavAPI()
        asset_details = api.asset_details(asset_id=14355608)
        ip = asset_details.address
        self.assertEqual(ip, ipaddress.ip_address("172.16.76.84"))

    def test_gav_all_asset_details(self):
        api = gav.GavAPI()
        assets, has_more, last_seen_asset_id = api.all_asset_details()
        self.assertTrue(len(assets) == 100)

    def test_gav_load_all_asset_details(self):
        api = gav.AllAssetDetailsORM()
        api.load()
        with orm.Session(api.engine) as session:
            host_list = (
                session.query(asset_details_orm.AssetItem)
                .where(asset_details_orm.AssetItem.asset_id == 14355608)
                .all()
            )
            self.assertEqual(host_list[0].ip, ipaddress.ip_address("172.16.76.84"))



"""
Qualys VMDR API Module

Typical usage example:
vmdr_orm = vmdr.HostListDetectionORM()
vmdr_orm.load()
session = orm.Session(vmdr_orm.engine):
host_list = session.query(HostORM).all()
stmt = session.query(HostORM)
hosts = vmdr_orm.query(stmt)[0]
host = hosts.host[0]
session.close()
"""

import os
import re
from typing import Any
import zoneinfo

import sqlalchemy.orm as orm
from xsdata.formats.dataclass.parsers import XmlParser

from . import URLS, qutils
from .base import QualysAPIBase, QualysORMMixin
from .models.vmdr import (
    host_list_vm_detection_orm,
    host_list_vm_detection_output,
    knowledge_base_vuln_list_output,
)
from .models.vmdr import host_list_output, host_list_orm


class VmdrAPI(QualysAPIBase):
    """Qualys VMDR API Class.  Contains methods for interacting with the VMDR API."""

    def __init__(
        self,
        config_file: str = str(os.path.join(os.path.expanduser("~"), ".qualyspy")),
        x_requested_with: str = "QualysPy Python Library",
    ) -> None:
        """Initialize the VMDR API Class.

        Args:
            config_file (str, optional): Path to config file. Defaults to ~/etc/qualysby/config.ini.
            x_requested_with (str, optional): Value for X-Requested-With header. Defaults to
                "QualysPy Python Library".
        """
        super().__init__(config_file, x_requested_with)
        self.xmlparser = XmlParser()

    def host_list_detection(
        self,
        *,
        ids: int | list[int] | None = None,
        truncation_limit: int | None = None,
        id_min: int | None = None,
        qids: int | list[int] | None = None,
        show_qds: bool | None = None,
        qds_min: int | None = None,
        qds_max: int | None = None,
    ) -> tuple[host_list_vm_detection_output.HostList, bool, int]:
        """Get a list of hosts with associated vulnerability detections from the VMDR API.  A value
        of None for the parameters will use their default values in the API.

        Args:
            ids (int | list[int] | None, optional): Host IDs to query. Defaults to None.
            truncation_limit (int | None, optional): Maximum number of hosts to return. Defaults to
                None.
            id_min (int | None, optional): Minimum host list ID to return. Defaults to None.

        Returns:
            tuple[host_list_vm_detection_output.HostList, bool, int]: A tuple containing the
                host_list_vm_detection_output.HostList object, a boolean indicating whether the
                results were truncated, and the next id_min to use for the next call.
        """

        params = {
            "ids": ids,
            "truncation_limit": truncation_limit,
            "id_min": id_min,
            "qids": qids,
            "show_qds": show_qds,
            "qds_min": qds_min,
            "qds_max": qds_max,
        }
        cleaned_params = qutils.clean_dict(params)
        cleaned_params["action"] = "list"

        truncated = False
        next_id_min = 0

        response = self.get(URLS.host_list_detection, params=cleaned_params).text
        parsed: host_list_vm_detection_output.HostListVmDetectionOutput = (
            self.xmlparser.from_string(
                response, host_list_vm_detection_output.HostListVmDetectionOutput
            )
        )
        if parsed.response is None:
            raise ValueError("API call returned no response.")
        ret = parsed.response.host_list
        if ret is None:
            raise ValueError("API call returned no host_list.")
        if (
            parsed.response.warning is not None
            and parsed.response.warning.url is not None
        ):
            next_id_match = re.search(r"id_min=(\d+)", parsed.response.warning.url)
            if next_id_match is None:
                raise ValueError(
                    "Unable to parse URL in warning message. No id_min found.\n"
                    f"{parsed.response.warning}"
                )
            truncated = True
            next_id_min = int(next_id_match.group(1))

        return ret, truncated, next_id_min

    def host_list(
        self,
        *,
        ids: int | list[int] | None = None,
        truncation_limit: int | None = None,
        id_min: int | None = None,
    ) -> tuple[host_list_output.HostList, bool, int]:
        """Get a list of hosts from the VMDR API.  A value of None for the parameters will use their
            default values in the API.

        Args:
            ids (int | list[int] | None, optional): Host IDs to query. Defaults to None.

        Returns:
            tuple[host_list_vm_detection_output.HostList, bool, int]: A tuple containing the
                host_list_vm_detection_output.HostList object, a boolean indicating whether the
                results were truncated, and the next id_min to use for the next call.
        """

        params = {
            k: str(v) for k, v in locals().items() if (k != "self" and v is not None)
        }
        params["action"] = "list"

        truncated = False
        next_id_min = 0

        response = self.get(URLS.host_list, params=params).text
        parsed: host_list_output.HostListOutput = self.xmlparser.from_string(
            response, host_list_output.HostListOutput
        )
        if parsed.response is None:
            raise ValueError("API call returned no response.")
        ret = parsed.response.host_list
        if ret is None:
            raise ValueError("API call returned no host_list.")
        if (
            parsed.response.warning is not None
            and parsed.response.warning.url is not None
        ):
            next_id_match = re.search(r"id_min=(\d+)", parsed.response.warning.url)
            if next_id_match is None:
                raise ValueError(
                    "Unable to parse URL in warning message. No id_min found.\n"
                    f"{parsed.response.warning}"
                )
            truncated = True
            next_id_min = int(next_id_match.group(1))

        return ret, truncated, next_id_min

    def knowledgebase(
        self, *, details: str | None = None, ids: int | list[int] | None = None
    ) -> knowledge_base_vuln_list_output.VulnList:
        params = {"details": details, "ids": ids}
        params["action"] = "list"
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.get(URLS.knowledgebase, params=params_cleaned).text
        parsed: knowledge_base_vuln_list_output.KnowledgeBaseVulnListOutput = (
            self.xmlparser.from_string(
                raw_response,
                knowledge_base_vuln_list_output.KnowledgeBaseVulnListOutput,
            )
        )
        if parsed.response is not None and parsed.response.vuln_list is not None:
            for vuln in parsed.response.vuln_list.vuln:
                if vuln.published_datetime is not None:
                    vuln.published_datetime = vuln.published_datetime.replace(
                        tzinfo=zoneinfo.ZoneInfo("UTC")
                    )
            return parsed.response.vuln_list
        else:
            return knowledge_base_vuln_list_output.VulnList()


class HostListDetectionORM(VmdrAPI, QualysORMMixin):
    """Qualys VMDR Host List Detection ORM Class.  Contains methods for loading host
    detections into an ORM database.
    """

    def __init__(self, echo: bool = False) -> None:
        """Initialize the Host List Detection ORM Class.

        Args:
            echo (bool, optional): Whether to echo SQL statements. Defaults to False.
        """
        VmdrAPI.__init__(self)
        self.orm_base = host_list_vm_detection_orm.Base  # type: ignore
        QualysORMMixin.__init__(self, self, echo=echo)

    def _load(self, load_func: Any, **kwargs: Any) -> None:
        """Load host detections into the ORM database.

        Args:
            load_func (Any): Function to call to get host detections.
            **kwargs (Any): Keyword arguments to pass to load_func.
        """

        def load_set(to_load: list[host_list_vm_detection_orm.Host]) -> None:
            """Load a single set of host detections into the ORM database.

            Args:
                to_load (list[host_list_vm_detection_orm.Host]): List of host list detections to
                    load.
            """
            with orm.Session(self.engine) as session:
                session.add_all(to_load)
                for obj in to_load:
                    session.merge(obj)
                session.commit()

        kwargs.setdefault("truncation_limit", 10000)
        truncated = True
        next_id_min = None
        while truncated:
            kwargs["id_min"] = next_id_min
            to_load, truncated, next_id_min = load_func(**kwargs)
            to_load = [
                qutils.to_orm_object(obj, host_list_vm_detection_orm.Host)
                for obj in to_load.host
            ]
            load_set(to_load)

    def load(self, **kwargs: Any) -> None:
        """Load host detections into the ORM database.

        Args:
            **kwargs (Any): Keyword arguments to pass to VmdrAPI.host_list_detection().
        """

        self.safe_load(self._load, self.host_list_detection, **kwargs)


class HostListORM(VmdrAPI, QualysORMMixin):
    """Qualys VMDR Host List ORM Class.  Contains methods for loading hosts into an ORM database."""

    def __init__(self, echo: bool = False) -> None:
        """Initialize the Host List ORM Class.

        Args:
            echo (bool, optional): Whether to echo SQL statements. Defaults to False.
        """
        VmdrAPI.__init__(self)
        self.orm_base = host_list_orm.Base  # type: ignore
        QualysORMMixin.__init__(self, self, echo=echo)

    def _load(self, load_func: Any, **kwargs: Any) -> None:
        """Load hosts into the ORM database.

        Args:
            load_func (Any): Function to call to get hosts.
            **kwargs (Any): Keyword arguments to pass to load_func.
        """

        def load_set(to_load: list[host_list_orm.Host]) -> None:
            """Load a single set of hosts into the ORM database.

            Args:
                to_load (list[host_list_orm.Host]): List of hosts to load.
            """
            with orm.Session(self.engine) as session:
                session.add_all(to_load)
                for obj in to_load:
                    session.merge(obj)
                session.commit()

        kwargs.setdefault("truncation_limit", 10000)
        truncated = True
        next_id_min = None
        while truncated:
            kwargs["id_min"] = next_id_min
            to_load, truncated, next_id_min = load_func(**kwargs)
            to_load = [
                qutils.to_orm_object(obj, host_list_orm.Host) for obj in to_load.host
            ]
            load_set(to_load)

    def load(self, **kwargs: Any) -> None:
        """Load hosts into the ORM database.

        Args:
            **kwargs (Any): Keyword arguments to pass to VmdrAPI.host_list().
        """

        self.safe_load(self._load, self.host_list, **kwargs)



"""Data model for the Qualys VM Detection Output API.

The dataclasses in this module are generated from the Qualys DTD schema using xsdata.
"""

import datetime as dt
import ipaddress
from dataclasses import field
from typing import Any, List, Optional

from pydantic.dataclasses import dataclass
from xsdata.formats.converter import Converter, converter

from . import converters

DT_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


class _StrConverter(Converter):
    """Converter for str, as metadata is a reserved keyword in SQLAlchemy."""

    def deserialize(self, value: str, **kwargs: dict[Any, Any]) -> str:
        if value == "metadata":
            return "metadata_"
        return value

    def serialize(self, value: str, **kwargs: dict[Any, Any]) -> str:
        return str(value)


converter.register_converter(ipaddress.IPv4Address, converters.IPv4AddressConverter())
converter.register_converter(ipaddress.IPv6Address, converters.IPv6AddressConverter())
converter.register_converter(dt.timedelta, converters.TimeDeltaConverter())
converter.register_converter(str, _StrConverter())


@dataclass
class Attribute:
    class Meta:
        name = "ATTRIBUTE"

    name: Optional[str] = field(
        default=None,
        metadata={
            "name": "NAME",
            "type": "Element",
            "required": True,
        },
    )
    last_status: Optional[str] = field(
        default=None,
        metadata={
            "name": "LAST_STATUS",
            "type": "Element",
            "required": True,
        },
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "name": "VALUE",
            "type": "Element",
            "required": True,
        },
    )
    last_success_date: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_SUCCESS_DATE",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_error_date: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_ERROR_DATE",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_error: Optional[str] = field(
        default=None,
        metadata={
            "name": "LAST_ERROR",
            "type": "Element",
        },
    )


@dataclass
class CloudTag:
    class Meta:
        name = "CLOUD_TAG"

    name: Optional[str] = field(
        default=None,
        metadata={
            "name": "NAME",
            "type": "Element",
            "required": True,
        },
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "name": "VALUE",
            "type": "Element",
            "required": True,
        },
    )
    last_success_date: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_SUCCESS_DATE",
            "type": "Element",
            "required": True,
            "format": DT_FORMAT,
        },
    )


@dataclass
class DnsData:
    class Meta:
        name = "DNS_DATA"

    hostname: Optional[str] = field(
        default=None,
        metadata={
            "name": "HOSTNAME",
            "type": "Element",
        },
    )
    domain: Optional[str] = field(
        default=None,
        metadata={
            "name": "DOMAIN",
            "type": "Element",
        },
    )
    fqdn: Optional[str] = field(
        default=None,
        metadata={
            "name": "FQDN",
            "type": "Element",
        },
    )


@dataclass
class Param:
    class Meta:
        name = "PARAM"

    key: Optional[str] = field(
        default=None,
        metadata={
            "name": "KEY",
            "type": "Element",
            "required": True,
        },
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "name": "VALUE",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class Qds:
    class Meta:
        name = "QDS"

    severity: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    value: Optional[int] = field(
        default=None,
        metadata={
            "required": True,
        },
    )


@dataclass
class QdsFactor:
    class Meta:
        name = "QDS_FACTOR"

    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    value: Optional[int] = field(
        default=None,
        metadata={
            "required": True,
        },
    )


@dataclass
class Tag:
    class Meta:
        name = "TAG"

    tag_id: Optional[int] = field(
        default=None,
        metadata={
            "name": "TAG_ID",
            "type": "Element",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "name": "NAME",
            "type": "Element",
            "required": True,
        },
    )
    color: Optional[str] = field(
        default=None,
        metadata={
            "name": "COLOR",
            "type": "Element",
        },
    )
    background_color: Optional[str] = field(
        default=None,
        metadata={
            "name": "BACKGROUND_COLOR",
            "type": "Element",
        },
    )


@dataclass
class Warning:
    class Meta:
        name = "WARNING"

    code: Optional[str] = field(
        default=None,
        metadata={
            "name": "CODE",
            "type": "Element",
        },
    )
    text: Optional[str] = field(
        default=None,
        metadata={
            "name": "TEXT",
            "type": "Element",
            "required": True,
        },
    )
    url: Optional[str] = field(
        default=None,
        metadata={
            "name": "URL",
            "type": "Element",
        },
    )


@dataclass
class Azure:
    class Meta:
        name = "AZURE"

    attribute: List[Attribute] = field(
        default_factory=list,
        metadata={
            "name": "ATTRIBUTE",
            "type": "Element",
        },
    )


@dataclass
class CloudProviderTags:
    class Meta:
        name = "CLOUD_PROVIDER_TAGS"

    cloud_tag: List[CloudTag] = field(
        default_factory=list,
        metadata={
            "name": "CLOUD_TAG",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Ec2:
    class Meta:
        name = "EC2"

    attribute: List[Attribute] = field(
        default_factory=list,
        metadata={
            "name": "ATTRIBUTE",
            "type": "Element",
        },
    )


@dataclass
class Google:
    class Meta:
        name = "GOOGLE"

    attribute: List[Attribute] = field(
        default_factory=list,
        metadata={
            "name": "ATTRIBUTE",
            "type": "Element",
        },
    )


@dataclass
class ParamList:
    class Meta:
        name = "PARAM_LIST"

    param: List[Param] = field(
        default_factory=list,
        metadata={
            "name": "PARAM",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class QdsFactors:
    class Meta:
        name = "QDS_FACTORS"

    qds_factor: List[QdsFactor] = field(
        default_factory=list,
        metadata={
            "name": "QDS_FACTOR",
            "type": "Element",
        },
    )


@dataclass
class Tags:
    class Meta:
        name = "TAGS"

    tag: List[Tag] = field(
        default_factory=list,
        metadata={
            "name": "TAG",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Detection:
    class Meta:
        name = "DETECTION"

    qid: Optional[int] = field(
        default=None,
        metadata={
            "name": "QID",
            "type": "Element",
            "required": True,
        },
    )
    type: Optional[str] = field(
        default=None,
        metadata={
            "name": "TYPE",
            "type": "Element",
            "required": True,
        },
    )
    severity: Optional[int] = field(
        default=None,
        metadata={
            "name": "SEVERITY",
            "type": "Element",
        },
    )
    port: Optional[int] = field(
        default=None,
        metadata={
            "name": "PORT",
            "type": "Element",
        },
    )
    protocol: Optional[str] = field(
        default=None,
        metadata={
            "name": "PROTOCOL",
            "type": "Element",
        },
    )
    fqdn: Optional[str] = field(
        default=None,
        metadata={
            "name": "FQDN",
            "type": "Element",
        },
    )
    ssl: Optional[bool] = field(
        default=None,
        metadata={
            "name": "SSL",
            "type": "Element",
        },
    )
    instance: Optional[str] = field(
        default=None,
        metadata={
            "name": "INSTANCE",
            "type": "Element",
        },
    )
    results: Optional[str] = field(
        default=None,
        metadata={
            "name": "RESULTS",
            "type": "Element",
        },
    )
    status: Optional[str] = field(
        default=None,
        metadata={
            "name": "STATUS",
            "type": "Element",
        },
    )
    first_found_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "FIRST_FOUND_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_found_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_FOUND_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    qds: Optional[Qds] = field(
        default=None,
        metadata={
            "name": "QDS",
            "type": "Element",
        },
    )
    qds_factors: Optional[QdsFactors] = field(
        default=None,
        metadata={
            "name": "QDS_FACTORS",
            "type": "Element",
        },
    )
    times_found: Optional[int] = field(
        default=None,
        metadata={
            "name": "TIMES_FOUND",
            "type": "Element",
        },
    )
    last_test_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_TEST_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_update_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_UPDATE_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_fixed_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_FIXED_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    first_reopened_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "FIRST_REOPENED_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_reopened_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_REOPENED_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    times_reopened: Optional[int] = field(
        default=None,
        metadata={
            "name": "TIMES_REOPENED",
            "type": "Element",
        },
    )
    service: Optional[str] = field(
        default=None,
        metadata={
            "name": "SERVICE",
            "type": "Element",
        },
    )
    is_ignored: Optional[bool] = field(
        default=None,
        metadata={
            "name": "IS_IGNORED",
            "type": "Element",
        },
    )
    is_disabled: Optional[bool] = field(
        default=None,
        metadata={
            "name": "IS_DISABLED",
            "type": "Element",
        },
    )
    affect_running_kernel: Optional[bool] = field(
        default=None,
        metadata={
            "name": "AFFECT_RUNNING_KERNEL",
            "type": "Element",
        },
    )
    affect_running_service: Optional[bool] = field(
        default=None,
        metadata={
            "name": "AFFECT_RUNNING_SERVICE",
            "type": "Element",
        },
    )
    affect_exploitable_config: Optional[bool] = field(
        default=None,
        metadata={
            "name": "AFFECT_EXPLOITABLE_CONFIG",
            "type": "Element",
        },
    )
    last_processed_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_PROCESSED_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    asset_cve: Optional[str] = field(
        default=None,
        metadata={
            "name": "ASSET_CVE",
            "type": "Element",
        },
    )


@dataclass
class Metadata:
    class Meta:
        name = "METADATA"

    ec2: List[Ec2] = field(
        default_factory=list,
        metadata={
            "name": "EC2",
            "type": "Element",
        },
    )
    google: List[Google] = field(
        default_factory=list,
        metadata={
            "name": "GOOGLE",
            "type": "Element",
        },
    )
    azure: List[Azure] = field(
        default_factory=list,
        metadata={
            "name": "AZURE",
            "type": "Element",
        },
    )


@dataclass
class Request:
    class Meta:
        name = "REQUEST"

    datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "DATETIME",
            "type": "Element",
            "required": True,
            "format": DT_FORMAT,
        },
    )
    user_login: Optional[str] = field(
        default=None,
        metadata={
            "name": "USER_LOGIN",
            "type": "Element",
            "required": True,
        },
    )
    resource: Optional[str] = field(
        default=None,
        metadata={
            "name": "RESOURCE",
            "type": "Element",
            "required": True,
        },
    )
    param_list: Optional[ParamList] = field(
        default=None,
        metadata={
            "name": "PARAM_LIST",
            "type": "Element",
        },
    )
    post_data: Optional[str] = field(
        default=None,
        metadata={
            "name": "POST_DATA",
            "type": "Element",
        },
    )


@dataclass
class DetectionList:
    class Meta:
        name = "DETECTION_LIST"

    detection: List[Detection] = field(
        default_factory=list,
        metadata={
            "name": "DETECTION",
            "type": "Element",
            "min_occurs": 1,
        },
    )


@dataclass
class Host:
    class Meta:
        name = "HOST"

    id: Optional[int] = field(
        default=None,
        metadata={
            "name": "ID",
            "type": "Element",
            "required": True,
        },
    )
    asset_id: Optional[int] = field(
        default=None,
        metadata={
            "name": "ASSET_ID",
            "type": "Element",
        },
    )
    ip: Optional[ipaddress.IPv4Address] = field(
        default=None,
        metadata={
            "name": "IP",
            "type": "Element",
        },
    )
    ipv6: Optional[ipaddress.IPv6Address] = field(
        default=None,
        metadata={
            "name": "IPV6",
            "type": "Element",
        },
    )
    tracking_method: Optional[str] = field(
        default=None,
        metadata={
            "name": "TRACKING_METHOD",
            "type": "Element",
        },
    )
    network_id: Optional[int] = field(
        default=None,
        metadata={
            "name": "NETWORK_ID",
            "type": "Element",
        },
    )
    os: Optional[str] = field(
        default=None,
        metadata={
            "name": "OS",
            "type": "Element",
        },
    )
    os_cpe: Optional[str] = field(
        default=None,
        metadata={
            "name": "OS_CPE",
            "type": "Element",
        },
    )
    dns: Optional[str] = field(
        default=None,
        metadata={
            "name": "DNS",
            "type": "Element",
        },
    )
    dns_data: Optional[DnsData] = field(
        default=None,
        metadata={
            "name": "DNS_DATA",
            "type": "Element",
        },
    )
    cloud_provider: Optional[str] = field(
        default=None,
        metadata={
            "name": "CLOUD_PROVIDER",
            "type": "Element",
        },
    )
    cloud_service: Optional[str] = field(
        default=None,
        metadata={
            "name": "CLOUD_SERVICE",
            "type": "Element",
        },
    )
    cloud_resource_id: Optional[int] = field(
        default=None,
        metadata={
            "name": "CLOUD_RESOURCE_ID",
            "type": "Element",
        },
    )
    ec2_instance_id: Optional[int] = field(
        default=None,
        metadata={
            "name": "EC2_INSTANCE_ID",
            "type": "Element",
        },
    )
    netbios: Optional[str] = field(
        default=None,
        metadata={
            "name": "NETBIOS",
            "type": "Element",
        },
    )
    qg_hostid: Optional[str] = field(
        default=None,
        metadata={
            "name": "QG_HOSTID",
            "type": "Element",
        },
    )
    last_scan_datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_SCAN_DATETIME",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_vm_scanned_date: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_VM_SCANNED_DATE",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_vm_scanned_duration: Optional[dt.timedelta] = field(
        default=None,
        metadata={
            "name": "LAST_VM_SCANNED_DURATION",
            "type": "Element",
        },
    )
    last_vm_auth_scanned_date: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_VM_AUTH_SCANNED_DATE",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    last_vm_auth_scanned_duration: Optional[dt.timedelta] = field(
        default=None,
        metadata={
            "name": "LAST_VM_AUTH_SCANNED_DURATION",
            "type": "Element",
        },
    )
    last_pc_scanned_date: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "LAST_PC_SCANNED_DATE",
            "type": "Element",
            "format": DT_FORMAT,
        },
    )
    tags: Optional[Tags] = field(
        default=None,
        metadata={
            "name": "TAGS",
            "type": "Element",
        },
    )
    metadata_: Optional[Metadata] = field(
        default=None,
        metadata={
            "name": "METADATA",
            "type": "Element",
        },
    )
    cloud_provider_tags: Optional[CloudProviderTags] = field(
        default=None,
        metadata={
            "name": "CLOUD_PROVIDER_TAGS",
            "type": "Element",
        },
    )
    detection_list: Optional[DetectionList] = field(
        default=None,
        metadata={
            "name": "DETECTION_LIST",
            "type": "Element",
            "required": True,
        },
    )


@dataclass
class HostList:
    class Meta:
        name = "HOST_LIST"

    host: List[Host] = field(
        default_factory=list,
        metadata={
            "name": "HOST",
            "type": "Element",
            "min_occurs": 1,
        },
    )

    def __add__(self, other: "HostList") -> "HostList":
        self.host.extend(other.host)
        return self


@dataclass
class Response:
    class Meta:
        name = "RESPONSE"

    datetime: Optional[dt.datetime] = field(
        default=None,
        metadata={
            "name": "DATETIME",
            "type": "Element",
            "required": True,
            "format": DT_FORMAT,
        },
    )
    host_list: Optional[HostList] = field(
        default=None,
        metadata={
            "name": "HOST_LIST",
            "type": "Element",
        },
    )
    warning: Optional[Warning] = field(
        default=None,
        metadata={
            "name": "WARNING",
            "type": "Element",
        },
    )


@dataclass
class HostListVmDetectionOutput:
    class Meta:
        name = "HOST_LIST_VM_DETECTION_OUTPUT"

    request: Optional[Request] = field(
        default=None,
        metadata={
            "name": "REQUEST",
            "type": "Element",
        },
    )
    response: Optional[Response] = field(
        default=None,
        metadata={
            "name": "RESPONSE",
            "type": "Element",
            "required": True,
        },
    )
